// Copyright © Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.notification.provider.aws.utils;

import com.amazonaws.services.sqs.AmazonSQS;
import com.amazonaws.services.sqs.model.DeleteMessageBatchRequest;
import com.amazonaws.services.sqs.model.DeleteMessageBatchRequestEntry;
import com.amazonaws.services.sqs.model.DeleteMessageBatchResult;
import com.amazonaws.services.sqs.model.Message;
import com.amazonaws.services.sqs.model.MessageAttributeValue;
import com.amazonaws.services.sqs.model.ReceiveMessageRequest;
import com.amazonaws.services.sqs.model.SendMessageRequest;
import com.amazonaws.services.sqs.model.SendMessageResult;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Component
public class SQSUtils {

    public List<Message> getMessages(AmazonSQS sqsClient, String sqsQueueUrl, int maxBatchRequestCount, int maxMessageCount) {
        int numOfMessages = maxBatchRequestCount;
        List<Message> messages = new ArrayList<>();
        do {
            ReceiveMessageRequest receiveMessageRequest = new ReceiveMessageRequest(sqsQueueUrl);
            receiveMessageRequest.setMaxNumberOfMessages(numOfMessages);
            receiveMessageRequest.withMessageAttributeNames("All");
            receiveMessageRequest.withAttributeNames("All");
            List<Message> retrievedMessages = sqsClient.receiveMessage(receiveMessageRequest).getMessages();
            messages.addAll(retrievedMessages);
            numOfMessages = retrievedMessages.size();

        } while (messages.size() < maxMessageCount && numOfMessages > 0);

        return messages;
    }

    public int deleteMessages(List<Message> messages, String sqsQueueUrl, AmazonSQS sqsClient) {
        int deletedCount = 0;
        List<DeleteMessageBatchRequestEntry> deleteEntries = new ArrayList<>();
        for (Message message : messages) {
            deleteEntries.add(new DeleteMessageBatchRequestEntry(message.getMessageId(), message.getReceiptHandle()));
            if (deleteEntries.size() == 10) {
                DeleteMessageBatchResult deleteResult = sqsClient.deleteMessageBatch(new DeleteMessageBatchRequest(sqsQueueUrl, deleteEntries));
                deletedCount += deleteResult.getSuccessful().size();
                deleteEntries.clear();
            }
        }
        if (!deleteEntries.isEmpty()) {
            DeleteMessageBatchResult deleteResult = sqsClient.deleteMessageBatch(new DeleteMessageBatchRequest(sqsQueueUrl, deleteEntries));
            deletedCount += deleteResult.getSuccessful().size();
            deleteEntries.clear();
        }
        return deletedCount;
    }

    public List<SendMessageResult> sendMessagesToDeadLetterQueue(String deadLetterQueueUrl, List<Message> messages, AmazonSQS sqsClient) {
        return messages.stream().map(message -> sendMessageToDeadLetterQueue(deadLetterQueueUrl, message, sqsClient))
                .collect(Collectors.toList());

    }

    private SendMessageResult sendMessageToDeadLetterQueue(String deadLetterQueueUrl, Message message, AmazonSQS sqsClient) {

        Map<String, MessageAttributeValue> messageAttributes = message.getMessageAttributes();
        MessageAttributeValue exceptionAttribute = new MessageAttributeValue()
                .withDataType("String");

        exceptionAttribute.setStringValue("Exception message: missing dataPartitionId");
        messageAttributes.put("Exception", exceptionAttribute);
        SendMessageRequest send_msg_request = new SendMessageRequest()
                .withQueueUrl(deadLetterQueueUrl)
                .withMessageBody(message.getBody())
                .withMessageAttributes(messageAttributes);
        return sqsClient.sendMessage(send_msg_request);
    }

}