// Copyright © Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.notification.provider.aws.queue.impl;

import static org.junit.Assert.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.when;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.withSettings;

import com.amazonaws.services.dynamodbv2.datamodeling.PaginatedQueryList;
import com.amazonaws.services.sqs.AmazonSQS;
import com.amazonaws.services.sqs.model.Message;
import com.amazonaws.services.sqs.model.MessageAttributeValue;
import com.amazonaws.services.sqs.model.SendMessageResult;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedConstruction;
import org.mockito.Mockito;
import org.mockito.internal.stubbing.defaultanswers.ForwardsInvocations;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperFactory;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperV2;
import org.opengroup.osdu.core.aws.ssm.K8sLocalParameterProvider;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.notification.provider.aws.model.FailedNotificationDoc;
import org.opengroup.osdu.notification.provider.aws.queue.impl.NotificationQueueServiceImpl;
import org.opengroup.osdu.notification.provider.aws.repository.SubscriptionRepository;
import org.opengroup.osdu.notification.provider.aws.service.AwsNotificationHandler;
import org.powermock.reflect.Whitebox;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

@RunWith(MockitoJUnitRunner.class)
public class NotificationQueueServiceImplTest {

    @Mock
    private SubscriptionRepository subscriptionRepository;

    @Mock
    private DynamoDBQueryHelperFactory dynamoDBQueryHelperFactory;

    @Mock
    private AwsNotificationHandler notificationHandler;

    @Mock
    private DynamoDBQueryHelperV2 dynamoDBQueryHelper;

    @Mock
    private AmazonSQS sqsClient;

    @InjectMocks
    NotificationQueueServiceImpl notificationQueueService;

    private static MockedConstruction<K8sLocalParameterProvider> mockedConstruction;

    @BeforeClass
    public static void setup() {
        mockedConstruction = Mockito.mockConstruction(K8sLocalParameterProvider.class,
                                 (mock, context) -> {
                                     //implement initializer for mock. Set return value for object A mock methods
                                     when(mock.getParameterAsString("notification-sqs-url")).thenReturn(
                                         "test-sqs-url");
                                 });
    }

    @AfterClass
    public static void close(){
        mockedConstruction.close();
    }

    @Before
    public void initTest() {
        Whitebox.setInternalState(notificationQueueService, "region", "us-east-1");
        Whitebox.setInternalState(notificationQueueService, "failedNotificationTablePath", "failed-notificationTable-path");
        Whitebox.setInternalState(notificationQueueService, "sqsClient", sqsClient);

        Subscription subscription1 = new Subscription();
        subscription1.setName("My listener");
        subscription1.setTopic("records-changed");
        subscription1.setPushEndpoint("/api/test/subscriber");
        subscription1.setNotificationId("de-859ea6a6-eefa-4e30-ba38-13f8cd71360a");
        subscription1.setId("testSubscription1Id");

        when(subscriptionRepository.getAllSubscriptionsByDataPartition(anyString())).thenReturn(Arrays.asList(subscription1));
        when(dynamoDBQueryHelperFactory.getQueryHelperUsingSSM(anyString())).thenReturn(dynamoDBQueryHelper);
    }

    @Test
    public void processNotificationMessage_success() throws Exception {
        List<FailedNotificationDoc> list = new ArrayList<>();
        when(dynamoDBQueryHelper.queryByGSI(any(), any())).thenReturn(mock(PaginatedQueryList.class, withSettings().defaultAnswer(new ForwardsInvocations(list))));

        HttpResponse response = new HttpResponse();
        response.setResponseCode(200);
        when(notificationHandler.notifySubscriber(any(), any(), any())).thenReturn(response);

        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testDataPartition"));

        List<Message> responseMessageList = notificationQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals("testMessage1", responseMessageList.get(0).getMessageId());
    }

    @Test
    public void processNotificationMessage_noSubscribers_success() {
        when(subscriptionRepository.getAllSubscriptionsByDataPartition(anyString())).thenReturn(Collections.emptyList());
        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testDataPartition"));

        List<Message> responseMessageList = notificationQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals("testMessage1", responseMessageList.get(0).getMessageId());
    }

    @Test
    public void processNotificationMessage_exceptionGettingSubscription_NoMessagesProcessed() {
        when(subscriptionRepository.getAllSubscriptionsByDataPartition(anyString())).thenAnswer(t -> {
            throw new Exception("test Exception");
        });

        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testDataPartition"));

        List<Message> responseMessageList = notificationQueueService.processNotificationMessages(messageList);
        assertEquals(0, responseMessageList.size());
    }

    @Test
    public void processNotificationMessage_existingFailedMessages_sentToRetry() {
        FailedNotificationDoc failedDoc = new FailedNotificationDoc();
        failedDoc.setId("testId");
        List<FailedNotificationDoc> list = Arrays.asList(failedDoc);
        when(dynamoDBQueryHelper.queryByGSI(any(), any())).thenReturn(mock(PaginatedQueryList.class, withSettings().defaultAnswer(new ForwardsInvocations(list))));
        when(sqsClient.sendMessage(any())).thenReturn(new SendMessageResult());

        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testDataPartition"));

        List<Message> responseMessageList = notificationQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals("testMessage1", responseMessageList.get(0).getMessageId());
    }

    @Test
    public void processNotificationMessage_exceptionInFailedMessagesDBQuery_messagesSentToRetry() {
        FailedNotificationDoc failedDoc = new FailedNotificationDoc();
        failedDoc.setId("testId");
        List<FailedNotificationDoc> list = Collections.emptyList();
        when(dynamoDBQueryHelper.queryByGSI(any(), any())).thenAnswer(t -> {
            throw new Exception();
        });
        when(sqsClient.sendMessage(any())).thenReturn(new SendMessageResult());

        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testDataPartition"));

        List<Message> responseMessageList = notificationQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals("testMessage1", responseMessageList.get(0).getMessageId());
    }

    @Test
    public void processNotificationMessage_non200ResponseFromNotification_sentToRetry() throws Exception {
        FailedNotificationDoc failedDoc = new FailedNotificationDoc();
        failedDoc.setId("testId");
        List<FailedNotificationDoc> list = Collections.emptyList();
        when(dynamoDBQueryHelper.queryByGSI(any(), any())).thenReturn(mock(PaginatedQueryList.class, withSettings().defaultAnswer(new ForwardsInvocations(list))));
        when(sqsClient.sendMessage(any())).thenReturn(new SendMessageResult());

        HttpResponse response = new HttpResponse();
        response.setResponseCode(500);
        when(notificationHandler.notifySubscriber(any(), any(), any())).thenReturn(response);

        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testDataPartition"));

        List<Message> responseMessageList = notificationQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals("testMessage1", responseMessageList.get(0).getMessageId());
    }

    @Test
    public void processNotificationMessage_exceptionFromNotification_messagesSentToRetry() throws Exception {
        FailedNotificationDoc failedDoc = new FailedNotificationDoc();
        failedDoc.setId("testId");
        List<FailedNotificationDoc> list = Collections.emptyList();
        when(dynamoDBQueryHelper.queryByGSI(any(), any())).thenReturn(mock(PaginatedQueryList.class, withSettings().defaultAnswer(new ForwardsInvocations(list))));
        when(sqsClient.sendMessage(any())).thenReturn(new SendMessageResult());

        when(notificationHandler.notifySubscriber(any(), any(), any())).thenThrow(new Exception());

        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testDataPartition"));

        List<Message> responseMessageList = notificationQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals("testMessage1", responseMessageList.get(0).getMessageId());
    }

    @Test
    public void processNotificationMessage_non200ResponseFromNotificationMultipleMessages_sentToRetry() throws Exception {
        FailedNotificationDoc failedDoc = new FailedNotificationDoc();
        failedDoc.setId("testId");
        List<FailedNotificationDoc> list = Collections.emptyList();
        when(dynamoDBQueryHelper.queryByGSI(any(), any())).thenReturn(mock(PaginatedQueryList.class, withSettings().defaultAnswer(new ForwardsInvocations(list))));
        when(sqsClient.sendMessage(any())).thenReturn(new SendMessageResult());

        HttpResponse response = new HttpResponse();
        response.setResponseCode(500);
        when(notificationHandler.notifySubscriber(any(), any(), any())).thenReturn(response);

        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testDataPartition"),
                                                  createMessage("testMessage2", "testDataPartition"));

        List<Message> responseMessageList = notificationQueueService.processNotificationMessages(messageList);
        assertEquals(2, responseMessageList.size());
        assertEquals("testMessage2", responseMessageList.get(1).getMessageId());
    }

    private Message createMessage(String messageId, String dataPartitionId) {
        Message message = new Message();
        message.setMessageId(messageId);
        MessageAttributeValue dataPartitionIdAttribute = new MessageAttributeValue()
            .withDataType("String");
        dataPartitionIdAttribute.setStringValue(dataPartitionId);
        message.getMessageAttributes().put("data-partition-id", dataPartitionIdAttribute);
        return message;
    }
}
