// Copyright © Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.notification.provider.aws.queue.impl;

import static org.junit.Assert.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.when;

import com.amazonaws.services.sqs.AmazonSQS;
import com.amazonaws.services.sqs.model.ChangeMessageVisibilityBatchRequestEntry;
import com.amazonaws.services.sqs.model.Message;
import com.amazonaws.services.sqs.model.MessageAttributeValue;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperFactory;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperV2;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.notification.provider.aws.model.FailedNotificationDoc;
import org.opengroup.osdu.notification.provider.aws.model.NotificationResult;
import org.opengroup.osdu.notification.provider.aws.model.RetryProcessResult;
import org.opengroup.osdu.notification.provider.aws.repository.SubscriptionRepository;
import org.opengroup.osdu.notification.provider.aws.service.AwsNotificationHandler;
import org.powermock.reflect.Whitebox;
import org.springframework.beans.factory.annotation.Value;

import java.util.Arrays;
import java.util.List;
import java.util.Optional;

@RunWith(MockitoJUnitRunner.class)
public class NotificationRetryQueueServiceTest {

    @Mock
    private AwsNotificationHandler notificationHandler;

    @Mock
    private DynamoDBQueryHelperFactory dynamoDBQueryHelperFactory;

    @Mock
    private SubscriptionRepository subscriptionRepository;

    @Value("${aws.dynamodb.failedNotificationTable.ssmPath}")
    private String failedNotificationTablePath;

    @InjectMocks
    private NotificationRetryQueueServiceImpl notificationRetryQueueService;

    @Mock
    private DynamoDBQueryHelperV2 dynamoDBQueryHelper;

    @Mock
    private AmazonSQS sqsClient;

    @Captor
    ArgumentCaptor<List<ChangeMessageVisibilityBatchRequestEntry>> changeMessageVisibilityRequestCaptor;

    @Before
    public void beforeTest() {
        Whitebox.setInternalState(notificationRetryQueueService, "failedNotificationTablePath", "failed-notificationTable-path");
        Whitebox.setInternalState(notificationRetryQueueService, "visibilityTimeout", 300);

        Subscription subscription1 = new Subscription();
        subscription1.setName("My listener");
        subscription1.setTopic("records-changed");
        subscription1.setPushEndpoint("/api/test/subscriber");
        subscription1.setNotificationId("de-859ea6a6-eefa-4e30-ba38-13f8cd71360a");
        subscription1.setId("testSubscription1Id");

        when(subscriptionRepository.getSubscriptionById(any(), any())).thenReturn(Optional.of(subscription1));
        when(dynamoDBQueryHelperFactory.getQueryHelperUsingSSM(anyString())).thenReturn(dynamoDBQueryHelper);
    }

    @Test
    public void processNotificationMessage_success() throws Exception {
        FailedNotificationDoc failedDoc = new FailedNotificationDoc();
        failedDoc.setId("testFailedRecordId");
        when(dynamoDBQueryHelper.loadByPrimaryKey(any(), any())).thenReturn(failedDoc);

        HttpResponse response = new HttpResponse();
        response.setResponseCode(200);
        when(notificationHandler.notifySubscriber(any(), any(), any())).thenReturn(response);

        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",1));

        List<RetryProcessResult> responseMessageList = notificationRetryQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals(NotificationResult.ACK, responseMessageList.get(0).getResult());
    }

    @Test
    public void processNotificationMessage_subscriptionNotFound_messagesTobeMarkedAcked() {
        when(subscriptionRepository.getSubscriptionById(any(), any())).thenReturn(Optional.empty());
        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",1));

        List<RetryProcessResult> responseMessageList = notificationRetryQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals(NotificationResult.ACK, responseMessageList.get(0).getResult());
    }

    @Test
    public void processNotificationMessage_generalException_messageNotProcessed() {
        when(subscriptionRepository.getSubscriptionById(any(), any())).thenAnswer(t -> {
            throw new Exception();
        });
        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",1));

        List<RetryProcessResult> responseMessageList = notificationRetryQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals(NotificationResult.NACK, responseMessageList.get(0).getResult());
    }

    @Test
    public void processNotificationMessage_non200ResponseFromSubscriber_messageNotAcked() throws Exception {
        HttpResponse response = new HttpResponse();
        response.setResponseCode(500);
        when(notificationHandler.notifySubscriber(any(), any(), any())).thenReturn(response);
        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId","testDataPartition", 1));
        List<RetryProcessResult> responseMessageList = notificationRetryQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals(NotificationResult.NACK, responseMessageList.get(0).getResult());
    }

    @Test
    public void processNotificationMessage_non200ResponseFromSubscriberMultipleMessages_messageNotAcked() throws Exception {
        HttpResponse response = new HttpResponse();
        response.setResponseCode(500);
        when(notificationHandler.notifySubscriber(any(), any(), any())).thenReturn(response);
        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId1","testDataPartition", 1),
                                                  createMessage("testMessage2", "testSubscription1Id", "testFailedRecordId2", "testDataPartition",1));
        List<RetryProcessResult> responseMessageList = notificationRetryQueueService.processNotificationMessages(messageList);
        assertEquals(2, responseMessageList.size());
        assertEquals(NotificationResult.NACK, responseMessageList.get(1).getResult());
    }

    @Test
    public void processNotificationMessage_exceptionSendingNotification_messageNotAcked() throws Exception {
        when(notificationHandler.notifySubscriber(any(), any(), any())).thenAnswer(t -> {
            throw new Exception("test Exception");
        });
        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",1));
        List<RetryProcessResult> responseMessageList = notificationRetryQueueService.processNotificationMessages(messageList);
        assertEquals(1, responseMessageList.size());
        assertEquals(NotificationResult.NACK, responseMessageList.get(0).getResult());
    }

    @Test
    public void changeMessageVisibilityTimeout_success() {
        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",1));
        notificationRetryQueueService.changeMessageVisibilityTimeout(sqsClient, "test-queue-url", messageList);
        Mockito.verify(sqsClient, times(1)).changeMessageVisibilityBatch(any(), changeMessageVisibilityRequestCaptor.capture());
        List<ChangeMessageVisibilityBatchRequestEntry> visibilityBatchRequestEntries = changeMessageVisibilityRequestCaptor.getValue();
        assertEquals(Integer.valueOf(300), visibilityBatchRequestEntries.get(0).getVisibilityTimeout());

        Mockito.reset(sqsClient);
        messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",15));
        notificationRetryQueueService.changeMessageVisibilityTimeout(sqsClient, "test-queue-url", messageList);
        Mockito.verify(sqsClient, times(1)).changeMessageVisibilityBatch(any(), changeMessageVisibilityRequestCaptor.capture());
        visibilityBatchRequestEntries = changeMessageVisibilityRequestCaptor.getValue();
        assertEquals(Integer.valueOf(600), visibilityBatchRequestEntries.get(0).getVisibilityTimeout());

        Mockito.reset(sqsClient);
        messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",25));
        notificationRetryQueueService.changeMessageVisibilityTimeout(sqsClient, "test-queue-url", messageList);
        Mockito.verify(sqsClient, times(1)).changeMessageVisibilityBatch(any(), changeMessageVisibilityRequestCaptor.capture());
        visibilityBatchRequestEntries = changeMessageVisibilityRequestCaptor.getValue();
        assertEquals(Integer.valueOf(1200), visibilityBatchRequestEntries.get(0).getVisibilityTimeout());

        Mockito.reset(sqsClient);
        messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",35));
        notificationRetryQueueService.changeMessageVisibilityTimeout(sqsClient, "test-queue-url", messageList);
        Mockito.verify(sqsClient, times(1)).changeMessageVisibilityBatch(any(), changeMessageVisibilityRequestCaptor.capture());
        visibilityBatchRequestEntries = changeMessageVisibilityRequestCaptor.getValue();
        assertEquals(Integer.valueOf(2400), visibilityBatchRequestEntries.get(0).getVisibilityTimeout());

        Mockito.reset(sqsClient);
        messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",45));
        notificationRetryQueueService.changeMessageVisibilityTimeout(sqsClient, "test-queue-url", messageList);
        Mockito.verify(sqsClient, times(1)).changeMessageVisibilityBatch(any(), changeMessageVisibilityRequestCaptor.capture());
        visibilityBatchRequestEntries = changeMessageVisibilityRequestCaptor.getValue();
        assertEquals(Integer.valueOf(4800), visibilityBatchRequestEntries.get(0).getVisibilityTimeout());

        Mockito.reset(sqsClient);
        messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId", "testDataPartition",55));
        notificationRetryQueueService.changeMessageVisibilityTimeout(sqsClient, "test-queue-url", messageList);
        Mockito.verify(sqsClient, times(1)).changeMessageVisibilityBatch(any(), changeMessageVisibilityRequestCaptor.capture());
        visibilityBatchRequestEntries = changeMessageVisibilityRequestCaptor.getValue();
        assertEquals(Integer.valueOf(9600), visibilityBatchRequestEntries.get(0).getVisibilityTimeout());

        Mockito.reset(sqsClient);
        messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId","testDataPartition", 61));
        notificationRetryQueueService.changeMessageVisibilityTimeout(sqsClient, "test-queue-url", messageList);
        Mockito.verify(sqsClient, times(1)).changeMessageVisibilityBatch(any(), changeMessageVisibilityRequestCaptor.capture());
        visibilityBatchRequestEntries = changeMessageVisibilityRequestCaptor.getValue();
        assertEquals(Integer.valueOf(43200), visibilityBatchRequestEntries.get(0).getVisibilityTimeout());
    }

    @Test
    public void changeMessageVisibilityTimeout_moreThan10messages_success() {
        List<Message> messageList = Arrays.asList(createMessage("testMessage1", "testSubscription1Id", "testFailedRecordId1", "testDataPartition", 1),
                                                  createMessage("testMessage2", "testSubscription1Id", "testFailedRecordId2", "testDataPartition", 1),
                                                  createMessage("testMessage3", "testSubscription1Id", "testFailedRecordId3", "testDataPartition", 1),
                                                  createMessage("testMessage4", "testSubscription1Id", "testFailedRecordId4", "testDataPartition", 1),
                                                  createMessage("testMessage5", "testSubscription1Id", "testFailedRecordId5", "testDataPartition", 1),
                                                  createMessage("testMessage6", "testSubscription1Id", "testFailedRecordId6", "testDataPartition", 1),
                                                  createMessage("testMessage7", "testSubscription1Id", "testFailedRecordId7", "testDataPartition", 1),
                                                  createMessage("testMessage8", "testSubscription1Id", "testFailedRecordId8", "testDataPartition", 1),
                                                  createMessage("testMessage9", "testSubscription1Id", "testFailedRecordId9", "testDataPartition", 1),
                                                  createMessage("testMessage10", "testSubscription1Id", "testFailedRecordId10", "testDataPartition", 1),
                                                  createMessage("testMessage11", "testSubscription1Id", "testFailedRecordId11", "testDataPartition", 1));
        notificationRetryQueueService.changeMessageVisibilityTimeout(sqsClient, "test-queue-url", messageList);
        Mockito.verify(sqsClient, times(2)).changeMessageVisibilityBatch(any(), any());

    }

    private Message createMessage(String messageId, String subscriptionId, String failedRecordId, String dataPartitionId, int receiveCount) {
        Message message = new Message();
        message.setMessageId(messageId);
        MessageAttributeValue subscriptionIdAttribute = new MessageAttributeValue()
            .withDataType("String");
        subscriptionIdAttribute.setStringValue(subscriptionId);
        message.getMessageAttributes().put("subscriptionId", subscriptionIdAttribute);

        MessageAttributeValue dataPartitionIdAttribute = new MessageAttributeValue()
            .withDataType("String");
        dataPartitionIdAttribute.setStringValue(dataPartitionId);

        MessageAttributeValue failedNotificationRecordId = new MessageAttributeValue()
            .withDataType("String");
        failedNotificationRecordId.setStringValue(failedRecordId);
        message.getMessageAttributes().put("failedNotificationRecordId", subscriptionIdAttribute);
        message.getMessageAttributes().put("data-partition-id", dataPartitionIdAttribute);

        message.getAttributes().put("ApproximateReceiveCount", String.valueOf(receiveCount));
        return message;
    }

}
