// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.cache;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.jupiter.api.Assertions;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.common.model.search.IdToken;
import org.opengroup.osdu.notification.provider.azure.cache.JwtCache;
import org.opengroup.osdu.notification.provider.azure.util.AppProperties;

import static org.junit.Assert.fail;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.initMocks;

@RunWith(MockitoJUnitRunner.class)
public class JWTCacheTest {
    private static String testTenantName = "testTenant";

    @Mock
    private AppProperties appProperties;

    @Mock
    private IdToken idToken;

    private JwtCache sut;

    @Before
    public void setup() {
        initMocks(this);
        when(appProperties.getMaxCacheSize()).thenReturn("10");
        sut = new JwtCache(appProperties);
    }

    @Test
    public void should_throwWhenCacheSizeIsInvalid_JwtCache(){
        // Set up
        when(appProperties.getMaxCacheSize()).thenReturn(null);
            try {

                // Act
                sut = new JwtCache(appProperties);

                // Assert
                fail("Should throw exception");
            } catch (Exception exception){
                Assertions.assertEquals(exception.getClass(), NumberFormatException.class);
        }

        // Set Up
        when(appProperties.getMaxCacheSize()).thenReturn("");
        try {

            // Act
            sut = new JwtCache(appProperties);

            // Assert
            fail("Should throw exception");
        } catch (Exception exception){
            Assertions.assertEquals(exception.getClass(), NumberFormatException.class);
        }

        // Set Up
        when(appProperties.getMaxCacheSize()).thenReturn("-1");
        try {

            // Act
            sut = new JwtCache(appProperties);

            // Assert
            fail("Should throw exception");
        } catch (Exception exception){
            Assertions.assertEquals(exception.getClass(), IllegalArgumentException.class);
        }
    }

    @Test
    public void should_return_getPut(){
        // Act
        IdToken observed = sut.get("test Tenant");

        // Assert
        Assert.assertNull(observed);

        // Set Up
        sut.put(testTenantName, idToken);

        // Act
        observed = sut.get(testTenantName);

        // Asset
        Assert.assertEquals(idToken, observed);

        // Act
        observed = sut.get("notInCache");

        // Assert
        Assert.assertNull(observed);

        try {
            // Act
            sut.put(null, null);
        } catch (Exception e){
            Assert.assertEquals(NullPointerException.class, e.getClass());
        }
    }

    @Test
    public void should_returnAppropriate_delete() {
        // Set Up
        sut.put(testTenantName, idToken);

        // Act
        sut.delete(testTenantName);

        // Assert
        IdToken observed = sut.get(testTenantName);
        Assert.assertNull(observed);
        try {

            // Act
            sut.delete("notInCache");

            // Assert
        } catch (Exception e){
            Assert.assertEquals(NullPointerException.class, e.getClass());
        }
    }

    @Test
    public void should_returnAppropriate_clearAll() {
        // Set Up
        sut.put(testTenantName, idToken);

        // Act
        sut.clearAll();

        // Assert
        IdToken observed = sut.get(testTenantName);
        Assert.assertNull(observed);
    }
}
