package org.opengroup.osdu.notification.provider.gcp.pubsub;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriver;
import org.opengroup.osdu.core.gcp.oqm.model.OqmDestination;
import org.opengroup.osdu.core.gcp.oqm.model.OqmSubscription;
import org.opengroup.osdu.core.gcp.oqm.model.OqmTopic;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Component;

import javax.validation.constraints.NotNull;
import java.util.List;

import static org.opengroup.osdu.notification.provider.gcp.pubsub.OqmSubscriberManager.*;

@Component
@RequiredArgsConstructor
@ConditionalOnProperty(name = "oqmDriver")
@Slf4j
public class MessageBrokerProvider {

    private final OqmDriver driver;
    private final OqmSubscriptionHandler subscriptionHandler;

    public void validateMessageBrokerConfiguration(TenantInfo tenantInfo) {
        List<OqmTopic> topics = subscriptionHandler.getTopics(tenantInfo.getDataPartitionId());
        topics.forEach(oqmTopic -> getMessageBrokerTopic(oqmTopic.getName(), tenantInfo));
        topics.forEach(
            oqmTopic -> getMessageBrokerTopic(oqmTopic.getName() + PUBLISH_SUFFIX, tenantInfo));
        // Source message broker subscriptions
        topics.forEach(oqmTopic -> getMessageBrokerSubscription(
            NOTIFICATION_PREFIX + oqmTopic.getName() + SERVICE_SUFFIX,
            tenantInfo));
        // Target message broker subscriptions for replication
        topics.forEach(oqmTopic -> getMessageBrokerSubscription(
            NOTIFICATION_PREFIX + oqmTopic.getName() + PUBLISH_SUFFIX,
            tenantInfo));
    }

    @NotNull
    public OqmSubscription getMessageBrokerSubscription(String subscriptionName, TenantInfo tenantInfo) {
        OqmSubscription subscription = driver.getSubscription(subscriptionName, getDestination(tenantInfo)).orElse(null);

        if (subscription == null) {
            log.error("Subscriber: {} not found. Create subscriber and restart service.",
                subscriptionName);
            throw new AppException(HttpStatus.INTERNAL_SERVER_ERROR.value(),
                "Required subscription not exists.",
                String.format(
                    "Required subscription not exists. Create subscription: %s and restart service.",
                    subscriptionName
                )
            );
        }
        return subscription;
    }

    @NotNull
    @Deprecated
    public OqmTopic getMessageBrokerTopic(String topicName, TenantInfo tenantInfo) {
        OqmTopic topic = driver.getTopic(topicName, getDestination(tenantInfo)).orElse(null);
        if (topic == null) {
            log.error("OQM: check for topic {} existence: ABSENT.", topicName);
            throw new AppException(
                HttpStatus.INTERNAL_SERVER_ERROR.value(),
                "Required topic not exists.",
                String.format(
                    "Required topic not exists. Create topic: %s and restart service.",
                    topicName
                )
            );
        }
        return topic;
    }

    private OqmDestination getDestination(TenantInfo tenantInfo) {
        return OqmDestination.builder()
                .partitionId(tenantInfo.getDataPartitionId())
                .build();
    }
}